/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_DESCRIPTOR_HXX_
# define PCAP_DESCRIPTOR_HXX_

# include <stdexcept>
# include <fstream>

# include <boost/lexical_cast.hpp>

# include "descriptor.hh"

# include <wipal/tool/endianness.hh>
# include <wipal/tool/exceptions.hh>

namespace pcapxx
{

  namespace internals
  {

    /// FIXME: TODO: Document.
    struct DummySetup
    {
      inline
      void
      operator () (std::streampos, std::streampos)
      {
      }
    };

    /// FIXME: TODO: Document.
    struct DummyUpdate
    {
      inline
      const std::string*
      operator () (std::streampos)
      {
	return 0;
      }
    };

  } // End of namespace pcapxx::internals.

  template <class Bottom>
  descriptor<Bottom>::descriptor(const std::string&	filename,
				 bool			build_index):
    super_type ()
  {
    internals::DummySetup	pi_setup;
    internals::DummyUpdate	pi_update;

    setup(filename, build_index, pi_setup, pi_update);
  }

  template <class Bottom>
  descriptor<Bottom>::descriptor(const std::string&	filename,
				 const index_type&	marks,
				 unsigned		pkt_count):
    super_type (),
    marks_ (marks),
    pkt_count_ (pkt_count)
  {
    internals::DummySetup	pi_setup;
    internals::DummyUpdate	pi_update;

    setup(filename, false, pi_setup, pi_update);
  }

  template <class Bottom>
  template <typename ProgressInfoSetupType,
	    typename ProgressInfoUpdateType>
  descriptor<Bottom>::descriptor(const std::string&		filename,
				 ProgressInfoSetupType&		pi_setup,
				 ProgressInfoUpdateType&	pi_update):
    super_type ()
  {
    setup(filename, true, pi_setup, pi_update);
  }

  template <class Bottom>
  typename descriptor<Bottom>::iterator
  descriptor<Bottom>::operator [] (size_t i)
  {
    assert(marks_.size() and "Cannot use operator [] without index.");

    const size_t	near_mark = i / mark_step;
    const size_t	extra_steps = i % mark_step;

    assert(i < pkt_count_);
    assert(near_mark < marks_.size());

    iterator		r (*this, marks_[near_mark], i - extra_steps);

    for (unsigned j = 0; j < extra_steps; ++j)
      {
	++r;
	assert(r != this->end());
      }
    return r;
  }

  template <class Bottom>
  size_t
  descriptor<Bottom>::size() const
  {
    assert(marks_.size() and "Cannot use size() without index.");

    return pkt_count_;
  }

  template <class Bottom>
  bool
  descriptor<Bottom>::swapped() const
  {
    return swap_;
  }

  template <class Bottom>
  int32_t
  descriptor<Bottom>::zone() const
  {
    return zone_;
  }

  template <class Bottom>
  int32_t
  descriptor<Bottom>::snaplen() const
  {
    return snaplen_;
  }

  template <class Bottom>
  typename descriptor<Bottom>::link_type
  descriptor<Bottom>::linktype() const
  {
    return type_;
  }

  template <class Bottom>
  void
  descriptor<Bottom>::expect(link_type l) const
  {
    using boost::lexical_cast;

    if (linktype() != l)
      throw std::invalid_argument (file_name() + ": Expected link type " +
				   lexical_cast<std::string>(l) +
				   ", got " +
				   lexical_cast<std::string> (linktype()));
  }

  template <class Bottom>
  std::streampos
  descriptor<Bottom>::file_size() const
  {
    return file_size_;
  }

  template <class Bottom>
  const std::string&
  descriptor<Bottom>::file_name() const
  {
    return file_name_;
  }

  template <class Bottom>
  dumper
  descriptor<Bottom>::dumper(const std::string& filename) const
  {
    return pcapxx::dumper (filename, int32_t (linktype()), snaplen());
  }

  template <class Bottom>
  template <typename ProgressInfoSetupType,
	    typename ProgressInfoUpdateType>
  void
  descriptor<Bottom>::setup(const std::string&		filename,
			    bool			build_index,
			    ProgressInfoSetupType&	pi_setup,
			    ProgressInfoUpdateType&	pi_update)
  {
    internals::file_header	h;
    std::ifstream		f (filename.c_str(), (std::ios::in |
						      std::ios::binary));

    if (not f)
      throw tool::file_error ("Open failed");
    file_name_ = filename;
    if (not f.seekg(0, std::ios::end))
      throw tool::seek_error ("Seek operation failed");
    file_size_ = f.tellg();
    if (not f.seekg(0, std::ios::beg))
      throw tool::seek_error ("Seek operation failed");

    if (not f.read(reinterpret_cast<char*> (&h), sizeof (h)) or
	f.gcount() != sizeof (h))
      throw tool::read_error ("Unreadable PCAP file header");
    switch (h.magic)
      {
      case 0xa1b2c3d4:
	swap_ = false;
	break;
      case 0xd4c3b2a1:
	swap_ = true;
	break;
      default:
	throw std::invalid_argument ("Not a PCAP trace");
      }

    if (tool::extract_short_u(h.version_major, swap_) !=
	internals::version_major)
      throw std::invalid_argument ("Invalid PCAP major version");
    if (tool::extract_short_u(h.version_minor, swap_) !=
	internals::version_minor)
      throw std::invalid_argument ("Invalid PCAP minor version");

    zone_	= tool::extract_long_s(h.thiszone, swap_);
    snaplen_	= tool::extract_long_u(h.snaplen, swap_);
    type_	= link_type (tool::extract_long_u(h.linktype, swap_));

    // Some traces advertised as Prism header do actually contain AVS
    // headers.  See
    // http://crawdad.cs.dartmouth.edu/wiki/pmwiki.php?n=Dataset.Ucsb-ietf2005
    if (type_ == PRISM_HEADER)
      {
	const const_iterator	b = this->begin();

	if (b != this->end())
	  {
	    const unsigned	magic =
	      tool::extract_big_endian_long_u(b->bytes().get());

	    if ((magic & 0xFFFFFF00) == 0x80211000)
	      type_ = IEEE802_11_RADIO_AVS;
	  }
      }

    pkt_count_ = 0;
    if (build_index)
      setup_marks(f, pi_setup, pi_update);
  }

  template <class Bottom>
  void
  descriptor<Bottom>::warn(const std::streampos& p, const std::string& msg)
  {
    std::cerr << file_name_ <<  ':' << p << ": WARNING: " << msg << '.'
	      << std::endl;
  }

  template <class Bottom>
  void
  descriptor<Bottom>::handle_truncation(const std::string& msg)
  {
    warn(file_size_, "corrupted file (" + msg + ')');
    if (not pkt_count_)
      throw std::invalid_argument ("Empty PCAP file");
    --pkt_count_;
  }

  template <class Bottom>
  template <typename ProgressInfoSetupType,
	    typename ProgressInfoUpdateType>
  void
  descriptor<Bottom>::setup_marks(std::istream&			f,
				  ProgressInfoSetupType&	pi_setup,
				  ProgressInfoUpdateType&	pi_update)
  {
# ifdef WP_TIME_PCAP_LOADING
    using namespace boost::posix_time;

    const ptime start = microsec_clock::universal_time();
# endif // WP_TIME_PCAP_LOADING

    {
      const std::streampos	first_packet_offset = f.tellg();

      marks_.push_back(first_packet_offset);
      pi_setup(first_packet_offset, file_size_);
    }

    internals::file_frame_header h;

    while (f.read(reinterpret_cast<char*> (&h), sizeof (h)))
      {
	if (f.gcount() != sizeof (h))
	  {
	    handle_truncation("truncated packet header");
	    break;
	  }

	char			junk[junk_len];

	h.caplen = tool::extract_long_u(h.caplen, swap_);
	while(h.caplen) // Skip packet data.
	  {
	    size_t		to_read = std::min(sizeof (junk),
						   size_t (h.caplen));

	    if (not f.read(junk, to_read))
	      {
		handle_truncation("truncated packet data");
		break;
	      }
	    h.caplen -= to_read;
	  }

	if (++pkt_count_ % mark_step == 0)
	  {
	    const std::streampos	current_offset = f.tellg();

	    marks_.push_back(current_offset);
	    if (const std::string* e = pi_update(current_offset))
	      throw tool::user_interruption (*e);
	  }
      }
    pi_update(file_size_);
    if (f.bad())
      throw tool::file_error ("Unrecoverable corruption in PCAP file");
    f.clear();

# ifdef WP_TIME_PCAP_LOADING
    std::cerr << (microsec_clock::universal_time() - start) << std::endl;
# endif // WP_TIME_PCAP_LOADING
  }

  inline
  void
  dumper::operator () (const pkthdr* h, const void* f)
  {
    // PCAP header.
    {
      internals::file_frame_header h_;

      h_.ts	= h->ts;
      h_.caplen	= h->caplen;
      h_.len	= h->len;

      if (not output_->write(reinterpret_cast<char*> (&h_), sizeof (h_)))
	throw tool::write_error("Could not write PCAP packet header");
    }

    // Packet.
    if (not output_->write(static_cast<const char*> (f), h->caplen))
      throw tool::write_error("Could not write packet data");

    if (++pkt_count_ % pcapxx::descriptor<tool::bottom>::mark_step == 0)
      marks_.push_back(output_->tellp());
  }

  inline
  descriptor<tool::bottom>
  dumper::descriptor() const
  {
    return pcapxx::descriptor<tool::bottom> (file_name_, marks_, pkt_count_);
  }


} // End of namespace pcapxx.

#endif // ! PCAP_DESCRIPTOR_HXX_
