/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_OPTIONS_HH_
# define TOOL_OPTIONS_HH_

# include <vector>
# include <string>
# include <set>
# include <utility>

# include <wipal/tool/endianness.hh>
# include <wipal/pcap/descriptor.hh>


/*----------.
| Options.  |
`----------*/

# define WP_OPT_NONE		""
# define WP_OPT_802CMP		"8"
# define WP_OPT_ATTR		"a:"
# define WP_OPT_BYTESCMP	"b"
# define WP_OPT_COLHEADS	"cC"
# define WP_OPT_DEFAULTCMP	"d"
# define WP_OPT_ERRS		"eE"
# define WP_OPT_HELP		"h"
# define WP_OPT_FRM_IDX		"iI"
# define WP_OPT_MAPPING		"m:"
# define WP_OPT_ENDIANNESS	"nN"
# define WP_OPT_PRISM_NOISE	"pP"
# define WP_OPT_QUIET		"q"
# define WP_OPT_IGNORE_REF	"r:"
# define WP_OPT_TIMECMP		"t"
# define WP_OPT_USEC		"uU"

// The underscore is needed to avoid clashes with automake.
# define WP_OPT_VERSION_	"v"

# define WP_OPT_LIST	(802CMP)(ATTR)(BYTESCMP)(COLHEADS)(DEFAULTCMP)	\
			(ERRS)(HELP)(FRM_IDX)(MAPPING)(ENDIANNESS)	\
			(PRISM_NOISE)(QUIET)(IGNORE_REF)(TIMECMP)(USEC)	\
			(VERSION_)

# define WP_HELP_STRING(L, M) "  -" #L "   " M "\n"

# define WP_HELP_802CMP							\
  WP_HELP_STRING(8, "Only compare 802.11 frames (no PHY header, no PCAP).")

# define WP_HELP_ATTR							\
  WP_HELP_STRING(a, "Attributes for unique frames identification "	\
		    "(default: " BOOST_PP_STRINGIZE(WP_DEFAULT_ATTRIBUTE) ").")

# define WP_HELP_BYTESCMP						\
  WP_HELP_STRING(b, "Only compare frame bytes (no PCAP header).")

# define WP_HELP_COLHEADS						\
  WP_HELP_STRING(c, "Do not print column headers "			\
                  "(default if output is not a tty).")			\
  WP_HELP_STRING(C, "Print column headers (default if output is a tty).")

# define WP_HELP_DEFAULTCMP						\
  WP_HELP_STRING(d, "Compare everything: PCAP data and bytes (default).")

# define WP_HELP_ERRS							\
  WP_HELP_STRING(e, "Do not print error values (default).")		\
  WP_HELP_STRING(E, "Print error values.")

# define WP_HELP_HELP							\
  WP_HELP_STRING(h, "Print this help.")

# define WP_HELP_FRM_IDX						\
  WP_HELP_STRING(i, "Do not print frame indices.")			\
  WP_HELP_STRING(I, "Print frame indices (default).")

# define WP_HELP_MAPPING						\
  WP_HELP_STRING(m, "MAC addresses mapping file "			\
	          "(default: " WP_MAPPING_DEFAULT ").")

# define WP_HELP_ENDIANNESS						\
  WP_HELP_STRING(n, "Force little endian for PHY headers.")		\
  WP_HELP_STRING(N, "Force big endian for PHY headers.")

# define WP_HELP_PRISM_NOISE						\
  WP_HELP_STRING(p, "Prism noise fields have no special meaning (default).") \
  WP_HELP_STRING(P, "Consider a non-zero Prism noise field as a PHY error.")

# define WP_HELP_QUIET							\
  WP_HELP_STRING(q, "Quiet ; produce minimal output (no column headers, etc.).")

# define WP_HELP_IGNORE_REF						\
  WP_HELP_STRING(r, "Ignore reference.  Blacklist a given x-y reference frame.")

# define WP_HELP_TIMECMP						\
  WP_HELP_STRING(t, "Compare 802.11 frames and time of arrival. "	\
		    "Use a precision of 106us.")

# define WP_HELP_USEC							\
  WP_HELP_STRING(u, "Do not print microseconds (default).")		\
  WP_HELP_STRING(U, "Print microseconds.")

# define WP_HELP_VERSION_						\
  WP_HELP_STRING(v, "Print version information.")


/*-------------.
| Attributes.  |
`-------------*/

# define WP_ATTR_HELP_hsh		"Hash frame using MD5 (16 bytes)."
# define WP_ATTR_HELP_tmp		"Use frame timestamp only (8 bytes)."
# define WP_ATTR_HELP_dst_tmp		"Use dst. + timestamp (12 bytes)."
# define WP_ATTR_HELP_src_tmp		"Use src. + timestamp (12 bytes)."
# define WP_ATTR_HELP_bss_tmp		"Use BSS + timestamp (12 bytes)."
# define WP_ATTR_HELP_src_bss_tmp	"Use src. + BSS + tmp. (16 bytes)."
# define WP_ATTR_HELP_seq_tmp		"Use SeqCtl + timestamp (10 bytes)."
# define WP_ATTR_HELP_seq_bss_tmp	"Use SC. + BSS + tmp. (14 bytes)."
# define WP_ATTR_HELP_seq_dst_bss_tmp	"Use SC. + dst. + BSS + tmp. "	\
					"(18 bytes)."
# define WP_ATTR_HELP_seq_src_bss_tmp	"Use SC. + src. + BSS + tmp. "	\
					"(18 bytes)."


/*--------------------.
| Other definitions.  |
`--------------------*/

# define WP_MAPPING_DEFAULT		"mapping"

namespace tool
{

  namespace opt
  {

    struct options
    {
      typedef std::vector<std::string>			input_type;
      typedef std::vector<std::string>			output_type;
      typedef std::set< std::pair<unsigned, unsigned> >	ref_blacklist_type;

      enum
	{
	  no_input			= 0,
	  no_output			= 0,
	  arbitrary_number_of_inputs	= 8,
	  arbitrary_number_of_outputs	= 8,
	};

      enum comp_method
	{
	  cmp_default,
	  cmp_802,
	  cmp_bytes,
	  cmp_time
	};

      input_type		input;
      output_type		output;
      std::string		mapping;
      std::string		attributes;
      endian::endianness	endianness;
      ref_blacklist_type	ref_blacklist;
      comp_method		cmp;
      bool			prism_noise_means_error;
      bool			print_column_headers;
      bool			print_errors;
      bool			print_frame_indices;
      bool			print_microseconds;
      bool			print_summary;
    };

    options	parse_args(int argc, char* const* argv,
			   const std::string& available_options,
			   int ninputs, int noutputs);

    template <class Fun>
    int
    dispatch_unique_id(const options& o, Fun f);

    void
    unsupported_linktype(const pcapxx::descriptor<>::link_type lt);

  } // End of namespace tool::opt.

} // End of namespace tool.

# include "options.hxx"

#endif // TOOL_OPTIONS_HH_
