/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_UNIQUE_ID_HH_
# define WIFI_FRAME_UNIQUE_ID_HH_

# include <ostream>
# include <boost/optional.hpp>

# include <wipal/pcap/frame_descriptor.hh>
# include <wipal/wifi/addr_mapping.hh>
# include <wipal/tool/hash.hh>

namespace wifi
{

  namespace frame
  {

    /// Internal compounds - not designed for users.
    namespace internals
    {
      struct tag1 {};
      struct tag2 {};
      struct tag3 {};
      struct tag4 {};

      template <class Tag>
      struct empty
      {
	bool operator  < (const empty& rhs) const;
	bool operator == (const empty& rhs) const;
	bool operator != (const empty& rhs) const;
      };

      template <class T, class Tag = tag1>
      struct adapter
      {
	adapter(const T& v = T());

	bool operator  < (const adapter& rhs) const;
	bool operator == (const adapter& rhs) const;
	bool operator != (const adapter& rhs) const;

	const T&	value() const;
	T&		value();

      private:
	T	v_;
      };

    } // End of namespace wifi::frame::internals.

    /**
     * Identifier for uniquely identifiable frames.
     *
     * A uniquely identifiable %frame is a %frame that is unique
     * inside a given trace.
     *
     * @todo Document.
     */
    template < class Id1 = internals::empty<internals::tag1>,
	       class Id2 = internals::empty<internals::tag2>,
	       class Id3 = internals::empty<internals::tag3>,
	       class Id4 = internals::empty<internals::tag4> >
    struct unique_id: private Id1, private Id2, private Id3, private Id4
    {
      /**
       * Construct a uniquely identifiable %frame from a %frame id
       * (from its @link pcapxx::pkthdr PCAP header @endlink) and some
       * distinctive attributes.
       *
       * The %frame id's only role is to help debugging.  It is only
       * used when dumping the %frame to a C++ stream.
       *
       * @see pcapxx::frame_descriptor
       */
      unique_id(unsigned	frame_id,
		const Id1&	id1 = Id1 (),
		const Id2&	id2 = Id2 (),
		const Id3&	id3 = Id3 (),
		const Id4&	id4 = Id4 ());

      /// @name Comparison operators
      //@{
      bool		operator < (const unique_id& rhs) const;
      bool		operator == (const unique_id& rhs) const;
      bool		operator != (const unique_id& rhs) const;
      //@}

      /// Dump a uniquely identifiable %frame onto a C++ stream.
      std::ostream&	print(std::ostream&) const;

      /// Get the frame id (for debugging purposse).
      unsigned		frame_id() const;

      /// @name Access to the distinctive attributes.
      //@{
      const Id1&	id1() const;
      const Id2&	id2() const;
      const Id3&	id3() const;
      const Id4&	id4() const;
      //@}

    private:
      unsigned		frame_id_;
    };

    /// Dump a @link uniquely_identifiable uniquely identifiable
    /// %frame @endlink onto a C++ stream.
    template <class Id1, class Id2, class Id3, class Id4>
    std::ostream&
    operator << (std::ostream&, const unique_id<Id1, Id2, Id3, Id4>&);

    /**
     * Unique identifier factory.
     *
     * Use the @c build() method to make unique identifiers of IEEE
     * 802.11 frames.
     *
     * This class is only @b forward declared (not declared, not
     * defined).  You must specialize it for each specific unique
     * identifier.
     */
    template <class UniqueId>
    struct unique_id_factory;

    namespace internals
    {

      template <class HeaderType>
      bool
      strip_phy_header(const void*&			frame,
		       size_t&				caplen,
		       const pcapxx::frame_descriptor& desc);

# define WP_UNIQUE_ID_DISSECT(F, D, HType, H)				\
      dissector<HType>	D;						\
      {									\
	const void*	frame;						\
	size_t		caplen;						\
									\
	if (not internals::strip_phy_header<HeaderType>(frame, caplen, F)) \
	  return boost::none_t ();					\
	D(frame, caplen);						\
      }									\
      const HType&	H = D.hooks()

    } // End of namespace wifi::frame::internals.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

/*---------------.
| Hash functions |
`---------------*/

WP_HASH_NAMESPACE_BEGIN

  template <class Tag>
  struct hash< wifi::frame::internals::empty<Tag> >
  {
    size_t
    operator () (const wifi::frame::internals::empty<Tag>&) const;
  };

  template <class T, class Tag>
  struct hash< wifi::frame::internals::adapter<T, Tag> >: public hash<T>
  {
    size_t
    operator () (const wifi::frame::internals::adapter<T, Tag>&) const;
  };

  template <class Id1, class Id2, class Id3, class Id4>
  struct hash< wifi::frame::unique_id<Id1, Id2, Id3, Id4> >:
      public hash<Id1>,
      public hash<Id2>,
      public hash<Id3>,
      public hash<Id4>
  {
    size_t
    operator () (const wifi::frame::unique_id<Id1, Id2, Id3, Id4>&) const;
  };

WP_HASH_NAMESPACE_END


# include "unique_id.hxx"

#endif // ! WIFI_FRAME_UNIQUE_ID_HH_
