/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_MACTIME_TRACKER_HXX_
# define WIFI_MACTIME_TRACKER_HXX_

# include "mactime_tracker.hh"

namespace wifi
{

  template <class Bottom>
  mactime_tracker<Bottom>::mactime_tracker(): low_bits_ (0),
					      wrap_count_ (0)
  {
  }

  template <class Bottom>
  void
  mactime_tracker<Bottom>::tick_32(uint32_t ts)
  {
    if (ts < low_bits_)
      ++wrap_count_;
    low_bits_ = ts;
  }

  template <class Bottom>
  void
  mactime_tracker<Bottom>::tick_64(uint64_t ts)
  {
    low_bits_ = uint32_t (ts);
    wrap_count_ = ts >> (sizeof (low_bits_) * 8);
  }

  template <class Bottom>
  void
  mactime_tracker<Bottom>::tick_tv(const struct timeval& tv)
  {
    const size_t		low_len = sizeof (low_bits_) * 8;
    const mpz_class		low_mask = uint32_t (-1);

    mpz_class		t (tv.tv_sec);
    t *= 1000000;
    t += tv.tv_usec;

    low_bits_ = mpz_class (t & low_mask).get_ui();
    wrap_count_ = mpz_class (t >> low_len).get_ui();
  }

  namespace internals
  {

    template <class I, class B>
    struct tick_switch
    {
      static void	tick(mactime_tracker<B>&		trk,
			     const pcapxx::frame_descriptor&	fd,
			     const I&				)
      {
	trk.tick_tv(fd.pcap_header()->ts);
      }
    };

    template <class B>
    struct tick_switch<uint32_t, B>
    {
      static void	tick(mactime_tracker<B>&		trk,
			     const pcapxx::frame_descriptor&	,
			     uint32_t				ts)
      {
	trk.tick_32(ts);
      }
    };

    template <class B>
    struct tick_switch<uint64_t, B>
    {
      static void	tick(mactime_tracker<B>&		trk,
			     const pcapxx::frame_descriptor&	,
			     uint64_t				ts)
      {
	trk.tick_64(ts);
      }
    };

    template <class B>
    struct tick_switch<struct timeval, B>
    {
      static void	tick(mactime_tracker<B>&		trk,
			     const pcapxx::frame_descriptor&	,
			     const struct timeval&		tv)
      {
	trk.tick_tv(tv);
      }
    };

  } // End of namespace wifi::internals.

  template <class Bottom>
  template <class HT>
  void
  mactime_tracker<Bottom>::tick(const pcapxx::frame_descriptor& frame,
				tool::endian::endianness	phy_end)
  {
    const pcapxx::pkthdr&	pcap = *frame.pcap_header();
    const bool			swap = tool::endian::need_swap(phy_end,
							       pcap.swapped);
    const HT* const		bytes = (reinterpret_cast<const HT*>
					 (frame.bytes().get()));

    if (pcap.caplen < bytes->len(pcap.caplen, swap))
      return;

    internals::tick_switch<typename HT::time_type::impl_type, Bottom>::
      tick(*this, frame, bytes->time_get(swap).get());
  }

  template <class Bottom>
  tool::microseconds
  mactime_tracker<Bottom>::microseconds() const
  {
    return tool::microseconds(wrap_count_, low_bits_);
  }

} // End of namespace wifi.

#endif // ! WIFI_MACTIME_TRACKER_HXX_
