/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <utility>
#include <stdexcept>
#include <iostream>
#include <cstdlib>
#include <map>

#include <boost/foreach.hpp>
#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include "config.hh"

#include <wipal/tool/options.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/frame/dissector/dissector.hh>
#include <wipal/wifi/frame/dissector/default_hooks.hh>
#include <wipal/wifi/frame/frame.hh>
#include <wipal/wifi/frame/mgt.hh>
#include <wipal/wifi/frame/filter/non_noisy_prism.hh>
#include <wipal/phy/empty.hh>
#include <wipal/phy/prism_header.hh>
#include <wipal/phy/rtap_header.hh>

namespace
{

  template <class PhyHeader>
  std::pair<size_t, const void* const>
  ieee80211(const pcapxx::frame_descriptor& d)
  {
    const PhyHeader* const	phy     = (reinterpret_cast<const PhyHeader*>
					   (d.bytes().get()));
    const pcapxx::pkthdr&	pcap    = *d.pcap_header();
    const bool			swapped = pcap.swapped;
    const size_t		caplen  = pcap.caplen;
    const size_t		len     = phy->len(caplen, swapped);

    return len > caplen?
      std::pair<size_t, const void* const> (0, 0):
      std::pair<size_t, const void* const> (caplen - len,
					    phy->decapsulate(caplen, swapped));
  }

  struct data_src: public wifi::frame::dissector_default_hooks
  {

    data_src(): check_ (false), src_ (0)
    {
    }

    void
    data_hook(const void* f, size_t)
    {
      // Ignore retransmitted frames.
      check_ = not wifi::frame::control_flag(3, f);
    }

    void
    addr_hook(const void*, size_t, unsigned i, const wifi::addr& a)
    {
      if (check_ and 2 == i)
	src_ = &a;
    }

    const wifi::addr*
    src() const
    {
      return src_;
    }

  private:
    bool		check_;
    const wifi::addr*	src_;

  }; // End of data_src.

  template <class PhyHeader, class T>
  bool
  find_data_dups(T& t, const tool::opt::options& o)
  {
    using wifi::frame::dissector;

    typedef pcapxx::frame_descriptor			frame_descriptor;
    typedef std::map<wifi::addr, frame_descriptor>	last_data_map;
    typedef typename T::const_iterator			const_iterator;
    typedef std::pair<size_t, const void*>		ieee80211_frame;

    last_data_map		map;
    bool			success (true);
    const const_iterator	e = t.end();

    for (const_iterator i = t.begin(); i != e; ++i)
      {
	const ieee80211_frame		frame = ieee80211<PhyHeader>(*i);
	const dissector<data_src>	dissect (frame.second, frame.first);

	if (dissect.src())
	  {
	    const std::pair<typename last_data_map::iterator, bool> r =
	      map.insert(std::make_pair(*dissect.src(), *i));

	    if (r.second)
	      continue;

	    const ieee80211_frame  last = ieee80211<PhyHeader>(r.first->second);

	    if (0 == memcmp(frame.second, last.second,
			    std::min(frame.first, last.first)))
	      {
		success = false;
		if (o.print_summary)
		  std::cout << *i << '\t' << r.first->second << '\n';
	      }
	    r.first->second = *i;
	  }
      }
    return success;
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  bool	success (true);

  try
    {
      const tool::opt::options	options =
	tool::opt::parse_args(argc, argv,
			      WP_OPT_COLHEADS
			      WP_OPT_PRISM_NOISE
			      WP_OPT_QUIET,
			      tool::opt::options::arbitrary_number_of_inputs,
			      0);

      BOOST_FOREACH(pcapxx::list_of_traces<> t, options.input)
	{
	  if (options.print_column_headers)
	    std::cout << t.file_name() << ":\n"
		         "Frame\tIs a duplicate of\n"
		         "=====\t=================" << std::endl;

	  if (t.empty())
	    continue;

	  if (options.prism_noise_means_error)
	    t.expect(pcapxx::descriptor<>::PRISM_HEADER);

	  switch (const pcapxx::descriptor<>::link_type lt =
		    t.iterable(0).linktype())
	    {

#define WP_case(R, Unused, LT)						\
	    case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	      if (options.prism_noise_means_error)			\
		{							\
		  typedef pcapxx::list_of_traces<>::const_iterator	i; \
		  typedef wifi::frame::filter::non_noisy_prism<i>	nnp; \
									\
		  nnp	n (t.begin(), t.end());				\
									\
		  success = find_data_dups<				\
		    BOOST_PP_TUPLE_ELEM(2, 1, LT)>(n, options) and success; \
		}							\
	      else							\
		success = find_data_dups<				\
	          BOOST_PP_TUPLE_ELEM(2, 1, LT)>(t, options) and success; \
	    break;

	      BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

	    default:
	      success = false;
	      tool::opt::unsupported_linktype(lt);
	      break;
	    }
	}
    }
  catch (std::exception& e)
    {
      success = false;
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
    }

  return success ? EXIT_SUCCESS : EXIT_FAILURE;
}
